/************************************************************************
 * This library is free software; you can redistribute it and/or	*
 * modify it under the terms of the GNU Library General Public		*
 * License as published by the Free Software Foundation; either		*
 * version 2 of the License, or (at your option) any later version.	*
 * 									*
 * This library is distributed in the hope that it will be useful,	*
 * but WITHOUT ANY WARRANTY; without even the implied warranty of	*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU	*
 * Library General Public License for more details.			*
 * 									*
 * You should have received a copy of the GNU Library General Public	*
 * License along with this library; if not, write to the		*
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,		*
 * Boston, MA  02111-1307, USA.						*
 * 									*
 * The GNU Library General Public License file is included in		*
 * Online-Documentation							*
 *									*
 * GNU's Homepage - http://www.gnu.ai.mit.edu/home.html			*
 *									*
 * Questions & comments referring to *FLIP++: please contact us:	*
 *									*
 * Snail-Mail:								*
 *	Dr. Wolfgang SLANY						*
 *	Database and Expert System Group (E184-2)     			*
 *	Institute of Information Systems               			*
 *	Vienna University of Technology					*
 *	Paniglgasse 16, A-1040 Vienna, AUSTRIA				*
 * Telephone:								*
 *	+43-1-58801-6123     (area-code 1 maybe changed to 201 in 1998) *
 * FAX:									*
 *	+43-1-5055304	     (area-code 1 maybe changed to 201 in 1998) *
 * eMail:								*
 *	sflip@dbai.tuwien.ac.at						*
 * WWW:									*
 *	http://www.dbai.tuwien.ac.at/proj/StarFLIP/			*
 *                                                                      *
 * $Id: hash.h,v 1.9 1997/06/24 17:33:15 reichel Exp $ *
 ***********************************************************************/


/* test */
/* =========================================================================
                          hash.h
   ========================================================================= */

#define MAX_HASH_SIZE     1000
#define OPT_HASH_SIZE     900
#define HASH_PREC         111111


class hash
{
  protected:
    
    float h_tab[MAX_HASH_SIZE];
    int items;
    int quater_it;
    char debug;
    int occupied;

    int h(plan *);
    int h_2(plan *);

  public:

    hash(int,char);
    ~hash(void);

    void init(void);
    int enter_hash(plan *);
    float get_hash(int);
    float delete_hash(int);
    int in_hash(plan *);
    int get_occupied(void);
};
    
hash::hash(int i,char d)
{
  debug = d;
  init();
  items = (i == -1) ? OPT_HASH_SIZE : i;
  quater_it = (int)(items / 4);
}

hash::~hash(void)
{
}

void hash::init(void)
{
  int i;
  
  if(debug)
    cout << "Init hashtable\n";
  for(i=0;i<MAX_HASH_SIZE;i++)
    h_tab[i] = -1;
  occupied = 0;
}

int hash::h(plan *p)
{
  int i,hlp;
  
  for(i=0;i<p->get_nr_of_entries()-5;i+=5)
    hlp = !(hlp & p->get_entry(i)) & !(hlp | p->get_entry(i));

  return hlp % items;
}

int hash::h_2(plan *p)
{
  long hlp = (long) (HASH_PREC * p->get_evaluation());
  return quater_it + (hlp % 9);
}

int hash::get_occupied(void)
{
  return occupied;
}

int hash::enter_hash(plan *p)
{
  int i=h(p),j=h_2(p),h;

  if(i==-1)
    return -1;
  else
  {
    h = i;
    do
    {
      if(h_tab[h%items] == -1)
      {
        h_tab[h%items] = p->get_evaluation();
        occupied++;
        return h+1;
      } else if(h_tab[h] == p->get_evaluation())
      {
        return h+1;
      }else
        h+=j;
    } while((h - items) < i); 
    h_tab[i] = p->get_evaluation();
    return i+1;
  }
}

float hash::get_hash(int i)
{
  if((i<0) || (i>items-1))
    return -1;
  else 
    return h_tab[i];
}

float hash::delete_hash(int i)
{
  float hlp;
  if((i<0) || (i>MAX_HASH_SIZE))
    return -1;
  else
  {
    hlp = h_tab[i];
    h_tab[i] = -1;
    if(hlp != -1) occupied--;
    return hlp;
  }
}

int hash::in_hash(plan *p)
{ 
  int i = h(p),j = h_2(p),h;

  if(i == -1)
    return -1;
  else
  {
    h = i;
    do
    {
      if(h_tab[h%items] == p->get_evaluation())
      {
        if(debug)
          cout << "Entry found in hash table\n";
        return (h%items)+1;
      } else
        h+=j;

    } while((h-items) < i);
    return 0;
  }
}


