/************************************************************************
 * This library is free software; you can redistribute it and/or	*
 * modify it under the terms of the GNU Library General Public		*
 * License as published by the Free Software Foundation; either		*
 * version 2 of the License, or (at your option) any later version.	*
 * 									*
 * This library is distributed in the hope that it will be useful,	*
 * but WITHOUT ANY WARRANTY; without even the implied warranty of	*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU	*
 * Library General Public License for more details.			*
 * 									*
 * You should have received a copy of the GNU Library General Public	*
 * License along with this library; if not, write to the		*
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,		*
 * Boston, MA  02111-1307, USA.						*
 * 									*
 * The GNU Library General Public License file is included in		*
 * Online-Documentation							*
 *									*
 * GNU's Homepage - http://www.gnu.ai.mit.edu/home.html			*
 *									*
 * Questions & comments referring to *FLIP++: please contact us:	*
 *									*
 * Snail-Mail:								*
 *	Dr. Wolfgang SLANY						*
 *	Database and Expert System Group (E184-2)     			*
 *	Institute of Information Systems               			*
 *	Vienna University of Technology					*
 *	Paniglgasse 16, A-1040 Vienna, AUSTRIA				*
 * Telephone:								*
 *	+43-1-58801-6123     (area-code 1 maybe changed to 201 in 1998) *
 * FAX:									*
 *	+43-1-5055304	     (area-code 1 maybe changed to 201 in 1998) *
 * eMail:								*
 *	sflip@dbai.tuwien.ac.at						*
 * WWW:									*
 *	http://www.dbai.tuwien.ac.at/proj/StarFLIP/			*
 *                                                                      *
 * $Id: plan.h,v 1.6 1997/06/24 17:33:16 reichel Exp $ *
 ***********************************************************************/


// $Log: plan.h,v $
// Revision 1.6  1997/06/24 17:33:16  reichel
//  ----------------------------------------------------------------------
// Address changed
//
//  Committing in optiflip//tabuflip
//
//  Modified Files:
//  	hash.h plan.h tabu.h tabudemo.cpp
//  ----------------------------------------------------------------------
//
// Revision 1.5  1997/06/24 13:26:02  reichel
//  ----------------------------------------------------------------------
// URL changed
//
//  Committing in optiflip//tabuflip
//
//  Modified Files:
//  	hash.h plan.h tabu.h tabudemo.cpp
//  ----------------------------------------------------------------------
//
// Revision 1.4  1997/05/26 13:24:15  reichel
//  ----------------------------------------------------------------------
// LGPL
//
//  Committing in sflip//conflip
//
//  Modified Files:
//  	bool.h cflp_dyn.cpp cflp_dyn.h cflp_flp.h cflperr.h
//  	cflplib.cpp cflplib.h cflplist.cpp cflplist.h common.cpp
//  	common.h conflip.cpp conflip.h demo1.cpp errort.cpp errort.h
//  	errtxt.cpp mytest.cpp
//  ----------------------------------------------------------------------
//  ----------------------------------------------------------------------
// LGPL
//
//  Committing in sflip//domflip
//
//  Modified Files:
//  	bool.h cflplist.cpp cflplist.h common.cpp common.h cviolat.cpp
//  	cviolat.h cviollst.cpp cviollst.h demo1.cpp demo2.cpp
//  	domagg.cpp domagg.h domaglst.cpp domaglst.h domain1.cpp
//  	domain1.h domerr.h domflip.cpp domflip.h domlib.cpp domlib.h
//  	domlib1.cpp domlib2.cpp domlist.cpp domlist.h errort.cpp
//  	errort.h errtxtdm.cpp genalg.cpp genalg.h hash.cpp hash.h
//  	order.cpp order.h repair.cpp repair.h replst.cpp replst.h
//  	repstepd.h tabu.cpp tabu.h
//  ----------------------------------------------------------------------
//  ----------------------------------------------------------------------
// LGPL
//
//  Committing in sflip//dynaflip
//
//  Modified Files:
//  	evaltree.cpp evaltree.h expint.cpp expint.h flip.h
//  	flipcvar.cpp flipcvar.h flipdemo.cpp flipoper.cpp flipoper.h
//  	flippara.cpp flippara.h fliprule.cpp fliprule.h fliptest.cpp
//  	inter.cpp
//  ----------------------------------------------------------------------
//  ----------------------------------------------------------------------
// LGPL
//
//  Committing in sflip//flip
//
//  Modified Files:
//  	flip.h flipcvar.cpp flipcvar.h flipdemo.cpp flipoper.cpp
//  	flipoper.h flippara.cpp flippara.h fliprule.cpp fliprule.h
//  	fliptest.cpp
//  ----------------------------------------------------------------------
//  ----------------------------------------------------------------------
// LGPL
//
//  Committing in sflip//optiflip/tabuflip
//
//  Modified Files:
//  	hash.h plan.h tabu.h tabudemo.cpp
//  ----------------------------------------------------------------------
//
// Revision 1.3  1996/02/14 10:24:30  raggl
// another test of logging
//
// Revision 1.2  1996/02/14 10:22:59  raggl
// test of logging


#define MAX_SCHED_ENTRIES     50
#define MAX_OUTLIST_ENTRIES   50
#define MAX_ADVICES           10

#define NR_OF_TEST_ENTRIES    50

#define A_NO_ADVICE           -1
#define A_END                  0
#define A_CHANGE               1
#define A_MOVE                 2
#define A_MOVE_GRP             3
#define A_INSERT               4
#define A_CHANGE_OL            5
#define A_REMOVE               6

#define MAX_REPAIRS  10


struct indextype
{
  int i_beg;
  int i_end;
};

struct advice
{
  char order[MAX_REPAIRS];
  indextype index;
};



class plan
{
  protected:

    int sched[MAX_SCHED_ENTRIES];
    int nr_of_entries;
    
    advice adv;
    float evaluation;
    void swap(int,int);

  public:

    plan(void);
    ~plan(void);

    int enter_sched(int,int);
    void copy_sched(plan *);
    void init_sched(void);
    int get_entry(int);

    void set_evaluation(float);
    float get_evaluation(void);

    void set_nr_of_entries(int);
    int get_nr_of_entries(void);

    advice get_advice(void);

    int call_repair(int,int,indextype);
    int change(int,int);
    int sng_move(int,int);
    int grp_move(int,int,int);
    int insert(int,int);
    int remove(int);

    double evaluate_plan(void);
    double test_evaluation(void);

    void make_advice(void);
    void test_advice(void);
    void std_advice(void);

    void random_it(void);
    void short_out_sched(void);
    void test_init();
};


plan::plan(void)
{
}

plan::~plan(void)
{
}

int plan::enter_sched(int pos,int nr)
{
  if((pos < 0) || (pos > MAX_SCHED_ENTRIES -1))
    return -1;
  else
  {
    sched[pos] = nr;
    return pos;
  }
}

void plan::copy_sched(plan *s)
{
  int i;
  for(i=0;i<MAX_SCHED_ENTRIES;i++)
    sched[i] = s->get_entry(i);
  evaluation    = s->get_evaluation();
  nr_of_entries = s->get_nr_of_entries();
}

void plan::init_sched(void)
{
  int i;
  for(i=0;i<MAX_SCHED_ENTRIES;i++)
    sched[i] = -1;
}

int plan::get_entry(int p)
{
  return sched[p];
}

void plan::set_evaluation(float b)
{
  evaluation = b;
}

float plan::get_evaluation(void)
{
  return evaluation;
}

void plan::set_nr_of_entries(int n)
{
  nr_of_entries = n;
}

int plan::get_nr_of_entries(void)
{
  return nr_of_entries;
}

advice plan::get_advice(void)
{
  return adv;
}
int plan::call_repair(int f,int pos ,indextype i)
{
  switch(f)
  {
    case A_MOVE:
      return sng_move(pos,i.i_beg);
    case A_CHANGE:
      return change(pos,i.i_beg);
    case A_MOVE_GRP:
      return grp_move(i.i_beg,i.i_end,pos);
  }
  return -1;   
}
void plan::swap(int s1,int s2)
{
  int help  = sched[s1];
  sched[s1] = sched[s2];
  sched[s2] = help;
}

int plan::change(int s1,int s2)
{
  if((s1 < 0) || (s2 > MAX_SCHED_ENTRIES) ||
     (s2 < 0) || (s2 > MAX_SCHED_ENTRIES))
    return -1;
  else if(s1 == s2)
    return 1;
  else
  {
    swap(s1,s2);
    return 1;
  }
}

int plan::sng_move(int von,int nach)
{
  int i;
  long h1,h2,h3;

  if((von<0) || (von>MAX_SCHED_ENTRIES) ||
     (nach<0) || (nach > MAX_SCHED_ENTRIES))
    return -1;
  else if(von==nach)
    return 1;
  h3 = sched[von];
  if(nach < von)
  {
    h1 = sched[nach];
    sched[nach] = h3; 
    for(i=nach+1;i<von + 1;i++)
    {
      h2 = sched[i];
      sched[i] = h1;
      h1 = h2;
    }
  } else
  {
    for(i=von;i<nach;i++)
      sched[i] = sched[i+1];    
    sched[nach] = h3;
  }
  return 1;
}


int plan::grp_move(int anf,int ende,int nach)
{
  int v;
  if(nach < anf)
    for(v=anf;v<=ende;v++)
      sng_move(v,nach);
  else if(nach > ende)
    for(v=ende;v>=anf;v--)
      sng_move(v,nach);
  return 1;
}

int plan::insert(int job,int nach)
{
  int i;
  if(nr_of_entries+1 < MAX_SCHED_ENTRIES)
  {
    for(i=nr_of_entries+1;i>nach+1;i--)
      sched[i] = sched[i-1];
    sched[nach+1] = job;
    nr_of_entries++;
    return nach+1;
  } else
    return -1;
}

int plan::remove(int s)        // wohin mit den entfernten job (outlist ?)
{
  int i;
  if((s<0) || (s>MAX_SCHED_ENTRIES))
    return -1;
  else
  {
    for(i=s;i<nr_of_entries-1;i++)
      sched[i] = sched[i+1];
    nr_of_entries--;
    return 1;
  }
}

/* -------------------------------
   Das mu/3 Witek machen !!!!!!!
   -------------------------------*/
double plan::evaluate_plan(void)
{
  evaluation = test_evaluation();
  return evaluation;
}
/* -------------------------------
   Das mu/3 Witek auch machen !!!!!!! 
   ------------------------------- */
void plan::make_advice()
{
  test_advice();
}

void plan::random_it(void)
{
  int i;
  for(i=0;(i<MAX_ADVICES) && (adv.order[0]>-1);i++)
  {
    adv.index.i_beg= rand() % nr_of_entries;
    adv.index.i_end = rand() % nr_of_entries;
  }
}
void plan::short_out_sched(void)
{
  int i;
  cout<<"\nSchedul:\n";
  for(i=0;i<nr_of_entries;i++)
    cout<<sched[i]<<" ";
  cout<<" Bewertung: "<<setprecision(4)<<evaluation;
  cout<<'\n';
}

void plan::test_init(void)
{
  int i;
  nr_of_entries = NR_OF_TEST_ENTRIES;
  for(i=0;i<nr_of_entries;i++)
    sched[i] = rand() % 100;
}

double plan::test_evaluation(void)
{
  int i,sum;
  sum =0;
  for(i=0;i<nr_of_entries-1;i++)
  {
    if(sched[i] > sched[i+1])
      sum += sched[i] - sched[i+1];
  }
  if(sum > 1000)
    return 0;
  else
    return (1-((double)sum/1000));
}

void plan::std_advice(void)
{
  adv.order[0] = A_CHANGE;
  adv.order[1] = A_MOVE;
  adv.order[2] = A_END;
  adv.index.i_beg = rand() % nr_of_entries;
}

void plan::test_advice(void)
{
  int i,pos1,max1=0,pos2,max2=0;
  for(i=0;i<nr_of_entries-1;i++)
    if((sched[i] - sched[i+1]) > max2)
    {
      max2 = sched[i] - sched[i+1];
      pos2 = i+1;
      if(max2 > max1)
      {
        int h;
        h = max2;max2 = max1;max1 = h;
        h = pos2;pos2=pos1;pos1 = h;
      }
    }
  adv.order[0] = A_CHANGE;
  adv.order[1] = A_MOVE;
  adv.order[2] = A_END;
  adv.index.i_beg = pos1;
}
