/************************************************************************
 * This library is free software; you can redistribute it and/or	*
 * modify it under the terms of the GNU Library General Public		*
 * License as published by the Free Software Foundation; either		*
 * version 2 of the License, or (at your option) any later version.	*
 * 									*
 * This library is distributed in the hope that it will be useful,	*
 * but WITHOUT ANY WARRANTY; without even the implied warranty of	*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU	*
 * Library General Public License for more details.			*
 * 									*
 * You should have received a copy of the GNU Library General Public	*
 * License along with this library; if not, write to the		*
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,		*
 * Boston, MA  02111-1307, USA.						*
 * 									*
 * The GNU Library General Public License file is included in		*
 * Online-Documentation							*
 *									*
 * GNU's Homepage - http://www.gnu.ai.mit.edu/home.html			*
 *									*
 * Questions & comments referring to *FLIP++: please contact us:	*
 *									*
 * Snail-Mail:								*
 *	Dr. Wolfgang SLANY						*
 *	Database and Expert System Group (E184-2)     			*
 *	Institute of Information Systems               			*
 *	Vienna University of Technology					*
 *	Paniglgasse 16, A-1040 Vienna, AUSTRIA				*
 * Telephone:								*
 *	+43-1-58801-6123     (area-code 1 maybe changed to 201 in 1998) *
 * FAX:									*
 *	+43-1-5055304	     (area-code 1 maybe changed to 201 in 1998) *
 * eMail:								*
 *	sflip@dbai.tuwien.ac.at						*
 * WWW:									*
 *	http://www.dbai.tuwien.ac.at/proj/StarFLIP/			*
 *                                                                      *
 * $Id: tabu.h,v 1.5 1997/06/24 17:33:17 reichel Exp $ *
 ***********************************************************************/




#include "hash.h"

#define DEF_SEARCH_DEPTH          40
#define DEF_SEARCH_REPEAT         3
#define DEF_TRIES_TILL_RANDOM    30

#define B_NOTHING                0
#define B_KEY                    1
#define B_TIME                   2
#define B_VALUE                  4

#define P_NOTHING                0
#define P_FAST                   1
#define P_OUTPUT                 2
#define P_DEBUG                  4
#define P_RANDOM                 8

struct break_struct
{
  char  what;
  float tm;
  float value;
};

struct para_struct
{
  int  SearchDepth;
  int  SearchRepeat;
  int  TriesTillRandom;
  int  HashSize;
  char what;
};

class tabu
{
  protected:

    hash *htab;
    plan *best_sched;

    int search_depth;
    int search_repeat;
    int tries_till_random;

    int deep;
    int another;

    char ready;
    char fast;
    char output;
    char debug;
    char rndm;

    time_t start_time;
    float  run_time;

    char bkey;
    float btime;
    float bvalue;

    void auswurf(float);
    int  breaking(float);
    void repair_step_deep(plan *);
    void repair_step_broad(plan *,int);
    
  public:

    tabu(plan *);
    tabu(plan *,break_struct,para_struct);
    ~tabu(void);

    void tabu_repair_deep(void);
    void tabu_repair_broad(void);
};



tabu::tabu(plan *s)
{
  htab          = new hash(OPT_HASH_SIZE,0);
  best_sched    = s;

  search_depth  = DEF_SEARCH_DEPTH;
  search_repeat = DEF_SEARCH_REPEAT;
   tries_till_random = DEF_TRIES_TILL_RANDOM;
  deep    = 0;
  another = 0;

  ready   = 0;
  fast    = 0;
  output  = 1;
  debug   = 0;
  rndm    = 1;

  bkey    = 0;
  btime   = 0;
  bvalue  = 0;

  start_time = time(NULL);
}

tabu::tabu(plan *s,break_struct b,para_struct p)
{
  best_sched    = s;
  search_depth  = (p.SearchDepth == -1) ? DEF_SEARCH_DEPTH : p.SearchDepth;
  search_repeat = (p.SearchRepeat == -1) ? DEF_SEARCH_REPEAT : p.SearchRepeat;
  tries_till_random = (p.TriesTillRandom == -1) ? DEF_TRIES_TILL_RANDOM : p.TriesTillRandom;
  deep    = 0;
  another = 0;

  ready = 0;
  if(p.what & P_FAST) fast = 1;
  else fast = 0;
  if(p.what & P_OUTPUT) output =1;
  else output = 0;
  if(p.what & P_DEBUG) debug  = 1;
  else debug = 0;
  if(p.what & P_RANDOM) rndm = 1;
  else rndm = 0;

  if(b.what & B_KEY)    bkey    = 1;
   else bkey    = 0;
  if(b.what & B_VALUE)  bvalue  = b.value;
  else bvalue  = 0;
  if(b.what & B_TIME)   btime   = b.tm;
  else btime   = 0;

  htab = new hash((((p.HashSize < 0) || (p.HashSize > MAX_HASH_SIZE))
                 ? OPT_HASH_SIZE : p.HashSize),debug);
  start_time = time(NULL);
}



tabu::~tabu()
{
  delete htab;
}


void tabu::auswurf(float b)
{
  cout << "\nbetter " << b;

  cout<<"\t"<<"time = "<<int(run_time);
  cout<<"\t"<<"repairstep = "<<deep;
  cout<<"\t"<<"repairtry = "<<another;
  cout<<"\n";
}

int tabu::breaking(float ev)
{
  run_time = double(time(NULL) - start_time);
  if(btime)
    if(btime<run_time)
      return 1;

  if(bvalue)
    if(bvalue<ev)
      return 1;
  
  return 0;
}

void tabu::tabu_repair_deep(void)
{
  plan try_sched,org_sched,b_sched;
  best_sched->evaluate_plan();
  try_sched.copy_sched(best_sched);
  b_sched.copy_sched(best_sched);
  org_sched.copy_sched(best_sched);
  
  htab->enter_hash(&try_sched);
  another = 1;ready = 0;
  while(!ready)
  {
    deep++;
    
    repair_step_deep(&try_sched);

    ready = breaking(try_sched.get_evaluation());

    if(try_sched.get_evaluation() > b_sched.get_evaluation())
    {
      if(output)
        auswurf(try_sched.get_evaluation());

      deep = 0;

      b_sched.copy_sched(&try_sched);

    } else
    {
      if(another<search_repeat)
      {
        another++;
        
        if(b_sched.get_evaluation() > best_sched->get_evaluation())
          best_sched->copy_sched(&b_sched);
        try_sched.copy_sched(&org_sched);
        b_sched.copy_sched(&org_sched);
        htab->enter_hash(&try_sched);
      } else
        ready = 1;
    }
  }
}

void tabu::tabu_repair_broad()
{
  plan try_sched,b_sched;
  char better = 0;
  best_sched->evaluate_plan();
  b_sched.copy_sched(best_sched);

  htab->enter_hash(&try_sched);
  another = 1;ready = 0;
  do
  {
    deep = 0;
    while((deep < search_repeat) && !ready)
    {
      deep++;

      try_sched.copy_sched(best_sched);
      repair_step_broad(&try_sched,better);
      ready = breaking(try_sched.get_evaluation());

      if(try_sched.get_evaluation() > b_sched.get_evaluation())
      {
        if(output)
          auswurf(try_sched.get_evaluation());
        b_sched.copy_sched(&try_sched);
      } 
    }
    another++;        
    best_sched->copy_sched(&b_sched);
    htab->enter_hash(best_sched);
  
  } while(another < search_depth);
  
}




void tabu::repair_step_deep(plan *b_sched)
{
  int tabu_verletzt = 0;
  plan try_sched;
  advice adv;
  int h,i;
  int found = 0;

  b_sched->evaluate_plan();
  try_sched.copy_sched(b_sched);
  try_sched.make_advice();


  adv = try_sched.get_advice();
   
  if(adv.order[0] == A_NO_ADVICE)
  {
    try_sched.std_advice();
    adv = try_sched.get_advice();
  } 
  for(i=0;i<MAX_REPAIRS;i++)
  {
    for(h=0;(h<b_sched->get_nr_of_entries()) && !found;h++)
    {
      if(h == adv.index.i_beg)
        h = h+1;
      try_sched.copy_sched(b_sched);
      try_sched.call_repair(adv.order[i],h,adv.index);
      if((try_sched.evaluate_plan() > b_sched->get_evaluation()) &&
         !(htab->in_hash(&try_sched)))
      {
        if(debug)
          cout << "Successfully modified\n";

        found=1; 
        htab->enter_hash(&try_sched);
      }
    }
  }
  if(found)
  {
    b_sched->copy_sched(&try_sched);
  }
}


void tabu::repair_step_broad(plan *b_sched,int better)
{
  int tabu_verletzt = 0;
  plan try_sched,best_sched;
  advice adv;
  int h,i;
  int found = 0;

  b_sched->evaluate_plan();
  try_sched.copy_sched(b_sched);
  best_sched.copy_sched(b_sched);

  try_sched.make_advice();

  if(better > tries_till_random)
  {
    try_sched.random_it();
    if(debug)
      cout<<"Random\n";
  }
 

  adv = try_sched.get_advice();
   
  if(adv.order[0] == A_NO_ADVICE)
  {
    try_sched.std_advice();
    adv = try_sched.get_advice();
  } 
  
  for(i=0; i<MAX_REPAIRS;i++)
  {
    for(h=0;(h<b_sched->get_nr_of_entries()) && !found;h++)
    {
      if(h == adv.index.i_beg)
        h = h + 1;
     
      try_sched.copy_sched(b_sched);   
      try_sched.call_repair(adv.order[i],h,adv.index);
      if((try_sched.evaluate_plan() > b_sched->get_evaluation()) &&
         !(htab->in_hash(&try_sched)))
      {
        if(debug)
          cout << "Successfully modified\n"; 
        best_sched.copy_sched(&try_sched);
        found=1;
        htab->enter_hash(&try_sched);
      }
    }
  }
   

  if(found)
  {
    b_sched->copy_sched(&try_sched);
  }
}












