/************************************************************************
 * This library is free software; you can redistribute it and/or	*
 * modify it under the terms of the GNU Library General Public		*
 * License as published by the Free Software Foundation; either		*
 * version 2 of the License, or (at your option) any later version.	*
 * 									*
 * This library is distributed in the hope that it will be useful,	*
 * but WITHOUT ANY WARRANTY; without even the implied warranty of	*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU	*
 * Library General Public License for more details.			*
 * 									*
 * You should have received a copy of the GNU Library General Public	*
 * License along with this library; if not, write to the		*
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,		*
 * Boston, MA  02111-1307, USA.						*
 * 									*
 * The GNU Library General Public License file is included in		*
 * Online-Documentation							*
 *									*
 * GNU's Homepage - http://www.gnu.ai.mit.edu/home.html			*
 *									*
 * Questions & comments referring to *FLIP++: please contact us:	*
 *									*
 * Snail-Mail:								*
 *	Dr. Wolfgang SLANY						*
 *	Database and Expert System Group (E184-2)     			*
 *	Institute of Information Systems               			*
 *	Vienna University of Technology					*
 *	Paniglgasse 16, A-1040 Vienna, AUSTRIA				*
 * Telephone:								*
 *	+43-1-58801-6123     (area-code 1 maybe changed to 201 in 1998) *
 * FAX:									*
 *	+43-1-5055304	     (area-code 1 maybe changed to 201 in 1998) *
 * eMail:								*
 *	sflip@dbai.tuwien.ac.at						*
 * WWW:									*
 *	http://www.dbai.tuwien.ac.at/proj/StarFLIP/			*
 *                                                                      *
 * $Id: tabu.cpp,v 1.4 1997/06/24 17:33:36 reichel Exp $ *
 ***********************************************************************/


#include "tabu.h"


Tabu::Tabu(Domain *s)
{
  htab          = new Hash(OPT_HASH_SIZE,0);
  best_sched    = s;

  search_depth  = DEF_SEARCH_DEPTH;
  search_repeat = DEF_SEARCH_REPEAT;
   tries_till_random = DEF_TRIES_TILL_RANDOM;
  deep    = 0;
  another = 0;

  ready   = 0;
  fast    = 0;
  output  = 1;
  debug   = 0;		
  rndm    = 1;

  bkey    = 0;
  btime   = 0;
  bvalue  = 0;

  start_time = time(NULL);
}

Tabu::Tabu(Domain *s,break_struct b,para_struct p)
{
  void *temp;
  ListOfAgg agglst;
  int pos;

  best_sched=s;		// get pointer only - it's not copy!
  // best_sched->copy(*s); 	// (real copy)


  search_depth  = (p.SearchDepth == -1) ? DEF_SEARCH_DEPTH : p.SearchDepth;
  search_repeat = (p.SearchRepeat == -1) ? DEF_SEARCH_REPEAT : p.SearchRepeat;
  tries_till_random = (p.TriesTillRandom == -1) ? DEF_TRIES_TILL_RANDOM : p.TriesTillRandom;
  deep    = 0;
  another = 0;

  // cout<<"search_depth: "<<search_depth<<endl;
  // cout<<"search_repeat: "<<search_repeat<<endl;
  // cout<<"tries_till_random: "<<tries_till_random<<endl;


  ready = 0;
  if(p.what & P_FAST) fast = 1;
  else fast = 0;
  if(p.what & P_OUTPUT) output =1;
  else output = 0;
  if(p.what & P_DEBUG) debug  = 1;
  else debug = 0;
  if(p.what & P_RANDOM) rndm = 1;
  else rndm = 0;

  if(b.what & B_KEY)    bkey    = 1;
   else bkey    = 0;
  if(b.what & B_VALUE)  bvalue  = b.value;
  else bvalue  = 0;
  if(b.what & B_TIME)   btime   = b.tm;
  else btime   = 0;

  htab = new Hash((((p.HashSize < 0) || (p.HashSize > MAX_HASH_SIZE))
                 ? OPT_HASH_SIZE : p.HashSize),debug);
  //cout<<"htab: "<< p.HashSize<<endl;

  start_time = time(NULL);	
  // cout<<"start_time: "<< start_time<<endl;

}

// ******************************************************************
Tabu::~Tabu()
{
  delete htab;
}

// ******************************************************************
void Tabu::auswurf(float b)
{
  cout << "\nbetter " << b;

  cout<<"\t"<<"time = "<<int(run_time);
  // cout<<"\t"<<"repairstep = "<<deep;
  // cout<<"\t"<<"repairtry = "<<another;
  cout<<"\n";
}
// ******************************************************************
int Tabu::breaking(float ev)
{
  run_time = double(time(NULL) - start_time);
  if(btime)
    if(btime<run_time)
      return 1;

  if(bvalue)
    if(bvalue<ev)
      return 1;
  
  return 0;
}
// ******************************************************************
void Tabu::tabu_repair_deep(void)
{
  Domain try_sched,org_sched,b_sched;

  best_sched->evaluate_plan();

  try_sched.copy(*best_sched);
  b_sched.copy(*best_sched);
  org_sched.copy(*best_sched);

  htab->enter_hash(&try_sched);
  another = 1;ready = 0;
  while(!ready)
  {
    deep++;
    
    repair_step_deep(&try_sched);

    ready = breaking(try_sched.get_evaluation());

    if(try_sched.get_evaluation() > b_sched.get_evaluation())
    {
      if(output)	
        auswurf(try_sched.get_evaluation());

      deep = 0;

      b_sched.copy(try_sched);

    } else
    {
      if(another<search_repeat)
      {
        another++;
        
        if(b_sched.get_evaluation() > best_sched->get_evaluation())
          best_sched->copy(b_sched);
        try_sched.copy(org_sched);
        b_sched.copy(org_sched);
        htab->enter_hash(&try_sched);
      } else
        ready = 1;
    }
  }
  // COPY	3.95 (for call tabu_repair_deep)
  // best_sched->copy(b_sched);

}
// ******************************************************************
void Tabu::tabu_repair_broad()
{
  Domain 	try_sched,b_sched;
  char 		better = 0;

  best_sched->evaluate_plan();
  b_sched.copy(*best_sched);


  htab->enter_hash(&try_sched);

  another = 1;ready = 0;
  do
  {
    deep = 0;
    while((deep < search_repeat) && !ready)	// 3 X
    {
      deep++;

      try_sched.copy(*best_sched);
      repair_step_broad(&try_sched,better);
      ready = breaking(try_sched.get_evaluation());

      if(try_sched.get_evaluation() > b_sched.get_evaluation())
      {
        if(output)		
          auswurf(try_sched.get_evaluation());
        b_sched.copy(try_sched);
      } 
    }
    another++;        
    best_sched->copy(b_sched);
    htab->enter_hash(best_sched);
  } while(another < search_depth);  
}

// ******************************************************************
void Tabu::repair_step_deep(Domain *b_sched)
{
  int 			tabu_verletzt = 0;
  Domain 		try_sched;
  SledForRepairs 	sfr;
  int 			h,i, tmp, rep_max;
  int 			found = 0;


  sfr = b_sched->get_repairl();
  rep_max = sfr.get_how_full();

  b_sched->evaluate_plan();
  try_sched.copy(*b_sched);
  
  for(h=1;(h<b_sched->get_nr_of_entries());h++)
  {
    for(i=1; ((i<rep_max)&& !found);i++)
    {     
      try_sched.copy(*b_sched);
      tmp=try_sched.get_aggs()->position();

      if( try_sched.get_advice(i,h))
      {
        if(try_sched.call_repair(h))
        {
  	  try_sched.get_aggs()->operator[](tmp);
          if((try_sched.evaluate_plan() > b_sched->get_evaluation()) &&
         !(htab->in_hash(&try_sched)))
          {
            if(debuginfo == 1)
              cout << "Successfully modified \n";
            found=1; 
            htab->enter_hash(&try_sched);
          }
        }
      }
      try_sched.get_aggs()->operator[](tmp);
    }
  }
  if(found)
  {
    b_sched->copy(try_sched);
  }
}

// ******************************************************************
void Tabu::repair_step_broad(Domain *b_sched,int better)
{
  int 			tabu_verletzt = 0;
  Domain 		try_sched,best_sched;
  SledForRepairs 	sfr;
  int 			h,i, tmp, rep, rep_max;
  int 			found = 0;
  char 			c;



  b_sched->evaluate_plan();
  try_sched.copy(*b_sched);
  best_sched.copy(*b_sched);


  sfr = try_sched.get_repairl();
  rep_max = sfr.get_how_full();

  if(better > tries_till_random)
  {
    try_sched.random_it();
    if(debuginfo == 1)
      cout<<"Random\n";
  }

  for(h=1;(h<b_sched->get_nr_of_entries());h++)
  {
   for(i=1; ((i<rep_max)&& !found);i++)
   {    
    try_sched.copy(*b_sched);  
    tmp=try_sched.get_aggs()->position();

    if( try_sched.get_advice(i,h))
    {
      if(try_sched.call_repair(h))
      {
	try_sched.get_aggs()->operator[](tmp);
      	if((try_sched.evaluate_plan() > b_sched->get_evaluation()) && !(htab->in_hash(&try_sched)))
        {
          if(debuginfo == 1)
            cout << "Successfully modified \n"; 
          best_sched.copy(try_sched);
          found=1;
          htab->enter_hash(&try_sched);
        }
      }
    }
    try_sched.get_aggs()->operator[](tmp);
   }
  }

  if(found)
  {
    b_sched->copy(try_sched);
  }
}
