/************************************************************************
 * This library is free software; you can redistribute it and/or	*
 * modify it under the terms of the GNU Library General Public		*
 * License as published by the Free Software Foundation; either		*
 * version 2 of the License, or (at your option) any later version.	*
 * 									*
 * This library is distributed in the hope that it will be useful,	*
 * but WITHOUT ANY WARRANTY; without even the implied warranty of	*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU	*
 * Library General Public License for more details.			*
 * 									*
 * You should have received a copy of the GNU Library General Public	*
 * License along with this library; if not, write to the		*
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,		*
 * Boston, MA  02111-1307, USA.						*
 * 									*
 * The GNU Library General Public License file is included in		*
 * Online-Documentation							*
 *									*
 * GNU's Homepage - http://www.gnu.ai.mit.edu/home.html			*
 *									*
 * Questions & comments referring to *FLIP++: please contact us:	*
 *									*
 * Snail-Mail:								*
 *	Dr. Wolfgang SLANY						*
 *	Database and Expert System Group (E184-2)     			*
 *	Institute of Information Systems               			*
 *	Vienna University of Technology					*
 *	Paniglgasse 16, A-1040 Vienna, AUSTRIA				*
 * Telephone:								*
 *	+43-1-58801-6123     (area-code 1 maybe changed to 201 in 1998) *
 * FAX:									*
 *	+43-1-5055304	     (area-code 1 maybe changed to 201 in 1998) *
 * eMail:								*
 *	sflip@dbai.tuwien.ac.at						*
 * WWW:									*
 *	http://www.dbai.tuwien.ac.at/proj/StarFLIP/			*
 *                                                                      *
 * $Id: tabu.h,v 1.4 1997/06/24 17:33:37 reichel Exp $ *
 ***********************************************************************/



#ifndef _TABU
#define _TABU

#include <time.h>
#include <stdlib.h>
#include <iomanip.h>

#include "hash.h"

#define DEF_SEARCH_DEPTH          6	// Org.: 40, 6
#define DEF_SEARCH_REPEAT         3
#define DEF_TRIES_TILL_RANDOM    30

#define B_NOTHING                0
#define B_KEY                    1
#define B_TIME                   2
#define B_VALUE                  4

#define P_NOTHING                0
#define P_FAST                   1
#define P_OUTPUT                 2
#define P_DEBUG                  4
#define P_RANDOM                 8

struct break_struct
{
  char  what;
  float tm;
  float value;
};

struct para_struct
{
  int  SearchDepth;
  int  SearchRepeat;
  int  TriesTillRandom;
  int  HashSize;
  char what;
};

//===========================================================================
class Tabu
//===========================================================================
{
  protected:

    Hash *htab;
    Domain *best_sched;

    int search_depth;
    int search_repeat;
    int tries_till_random;

    int deep;
    int another;

    char ready;
    char fast;
    char output;
    char debug;
    char rndm;

    time_t start_time;
    float  run_time;

    char bkey;
    float btime;
    float bvalue;

    void auswurf(float);
    int  breaking(float);
    void repair_step_deep(Domain *);
    void repair_step_broad(Domain *,int);
    
  public:

    Tabu(Domain *);
    Tabu(Domain *,break_struct,para_struct);
    ~Tabu(void);

    void tabu_repair_deep(void);
    void tabu_repair_broad(void);
};


   
#endif









